import requests

from clusterCore.exceptions import APIException404
from django.http import JsonResponse, HttpResponse, StreamingHttpResponse
import six, time, base64, hmac, hashlib, time
from urllib.parse import urlencode
from django.urls import reverse
from django.conf import settings
from django.utils import timezone
from django.utils.encoding import force_bytes, force_str
from django.contrib.sites.shortcuts import get_current_site
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated, AllowAny
from rest_framework import generics, status, filters
from rest_framework.decorators import api_view, permission_classes, authentication_classes
from django.utils.http import urlsafe_base64_encode, urlsafe_base64_decode
from rest_framework_simplejwt.views import TokenObtainPairView
from django.contrib.auth.models import User
from django.contrib import messages
from django.contrib.auth.models import Group
from django.contrib.auth import update_session_auth_hash
from django.contrib.auth.tokens import PasswordResetTokenGenerator
from types import SimpleNamespace
from clusterCore.tasks import generate_random_password, generate_human_readable_username, run_python_script, \
    test_proxmox_connection, connect_to_redis, create_elasticsearch_insecure_connection, get_random_available_ip, \
    check_space, generate_license_summary, is_valid_email, check_port_existence, send_mail, format_password, \
    generate_connection_string, checkPathExists, create_ssh_connection, create_mysql_connection, \
    get_postgres_connection, create_elasticsearch_insecure_connection, create_mongo_connection, \
    connect_to_oracle_database, connect_to_sqlserver_instance, create_winrm_connection, connect_to_sqlserver_instance, \
    find_tilda, remote_command, remote_command_win, drop_backup, check_host_connection, local_command
from clusterCore.exceptions import *
from itertools import chain
from psycopg2.extras import RealDictCursor
from django.template import Template, Context
from django.db.models import Exists, Count, IntegerField, OuterRef, Count, F, Value, Subquery
from django.db.models.functions import Cast, Coalesce
from django.db import connection
from rest_framework.authentication import BaseAuthentication
import json, os
from clusterCore.enums import DBType, IntegratedServiceTypes, ClusterRoles, Inventory, Status, Severity, Schedule, TaskState, ObjectType, Node_Size_Allowed
from clusterCore.models import (OracleSystemConfig, Profile, Organization, c123_host, SystemConfig, SystemSettings, UserPrivileges, c123_database, c123_node, c123_host_monitor,TaskLog,TaskStepLog,
ClusterSizing, CustomerKey, OracleClusterSizing, UserNotifications, Task, c123_inventory, IntegratedService, c123_cluster)
from .serializer import (ClusterSizingSerializer, OracleClusterSizingSerializer, OracleSystemConfigSerializer, SystemConfigSerializer, UserNotificationsSerializer, UserPrivilegesSerializer, UserSerializer, c123_DatabaseSerializer, c123_NodeSerializer,
                        c123_HostMonitorSerializer,
                        c123_AllNodeSerializer,
                        SelfTokenObtainPairSerializer,
                        c123_TaskLogSerializer,
                        c123_TaskStepLogSerializer,
                        c123_inventory_serializer,
                        c123_HostSerializer,
                        c123_TaskStepLogDetailSerializer,
                        c123_AddIntegratedServiceSerializer,
                        TaskSerializer, c123_ClusterStatusSerializer,
                        c123_TaskLogTextSerializer, c123_UserSerializer, ConfigSerializer, c123_IntegratedServiceSerializer,
                    object_to_json, format_size)

class SelfTokenObtainPairView(TokenObtainPairView):
    serializer_class = SelfTokenObtainPairSerializer
class TokenGenerator(PasswordResetTokenGenerator):
    def _make_hash_value(self, user, timestamp):
        return (
            six.text_type(user.pk) + six.text_type(timestamp) +
            six.text_type(user.is_active)
        )

@api_view(["POST"])
@permission_classes((IsAuthenticated, ))
def ValidateEmail(request):
    valid = is_valid_email(request.data.get("email"))
    return JsonResponse({"valid": valid}, status = 200)
@api_view(['GET'])
def getRoutes(request):
    routes = [
        '/api/token',
        '/api/token/refresh'
    ]
    return Response(routes,safe=False)

class NoAuthentication(BaseAuthentication):
    def authenticate(self, request):
        return None

class CurrentUserView(APIView):
    def get(self, request):
        current_user = request.user
        is_superuser = current_user.groups.filter(name='Administrators').exists() or current_user.is_superuser
        privileges, created = UserPrivileges.objects.get_or_create(user=current_user)
        serializer = c123_UserSerializer(current_user)
        privileges_serializer = UserPrivilegesSerializer(privileges)
        return JsonResponse({"data": serializer.data, "privileges": privileges_serializer.data, "is_superuser": is_superuser}, status = 200)
class UserTaskLogsByType(generics.GenericAPIView):
    permission_classes = [IsAuthenticated]
    queryset = User.objects.all()
    serializer_class = c123_TaskLogSerializer

    def get(self, request):
        response = {}
        queryset = self.get_queryset()
        response["start"] = TaskLog.objects.order_by("created_date").first().created_date
        response["end"] = TaskLog.objects.order_by("created_date").last().created_date
        for user in queryset:
            user_task_logs = TaskLog.objects.filter(created_by = user)
            task_types = user_task_logs.order_by().values('task').distinct()
            response[user.username] = {}
            for task_type in task_types:
                user_task_logs_by_type = self.serializer_class(user_task_logs.filter(task=task_type["task"]), many = True)
                response[user.username][task_type["task"]] = user_task_logs_by_type.data
        return Response(response, status=status.HTTP_200_OK)

class AddIntegratedService(generics.CreateAPIView):
    permission_classes = [IsAuthenticated]
    queryset = IntegratedService.objects.all()
    serializer_class = c123_AddIntegratedServiceSerializer
    def create(self, request, *args, **kwargs):
        super().create(request, *args, **kwargs)
        return JsonResponse({"detail": "Object created successfully."}, status=status.HTTP_201_CREATED)

class NodeList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_NodeSerializer
    def get_queryset(self):
        user = self.request.user
        if user.is_superuser or user.groups.filter(name='Administrators').exists():
            return c123_node.objects.filter(org_id = user.profile.org_id)
        return c123_node.objects.filter(org_id = user.profile.org_id, created_by = user)

    def list(self, request, *args, **kwargs):
        c123_node_objects = self.serializer_class(self.get_queryset(), many=True)
        if c123_node_objects:
            return Response(c123_node_objects.data)
        else:
            return Response({'Error':'There is no nodes'}, status=status.HTTP_404_NOT_FOUND)
class HostAvailableList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_HostSerializer
    def get_queryset(self):
        return c123_host.objects.filter(ind_deleted = False)

    def list(self, request, *args, **kwargs):
        c123_host_objects = self.serializer_class(self.get_queryset(), many=True)
        if c123_host_objects:
            return Response(c123_host_objects.data)
        else:
            return Response({'Error':'There is no available hosts'}, status=status.HTTP_404_NOT_FOUND)

class NodeAvailableList(generics.ListAPIView):
    permission_classes = [IsAuthenticated]
    serializer_class = c123_NodeSerializer

    def get_queryset(self):
        user = self.request.user

        # Subquery for counting slaves
        slaves_count_subquery = c123_node.objects.filter(
            type=DBType.REDIS.value,
            cluster_role=ClusterRoles.STANDBY.value,
            ind_deleted=False,
        ).annotate(
            slave_of_casted=Coalesce(Cast('ex_params__slave_of', output_field=IntegerField()), Value(0))  # Default 0 for None
        ).filter(
            slave_of_casted=F('id')  # Compare the casted value with the id
        ).values('id').annotate(count=Count('id')).values('count')[:1]  # Extract count value

        queryset = (
            c123_node.objects
            .select_related("host", "created_by", "last_update_by", "cluster")
            .annotate(
                slaves_count=Subquery(slaves_count_subquery, output_field=IntegerField())
            )
        )

        if user.is_superuser or user.groups.filter(name='Administrators').exists():
            return queryset.filter(
                ind_deleted=False,
                org_id=user.profile.org_id
            ).exclude(
                type=DBType.MONGO.value,
                cluster__ex_params__database_pwd__isnull=False
            ).order_by("-created_date")

        return queryset.filter(
            created_by=user,
            org_id=user.profile.org_id,
            ind_deleted=False
        ).exclude(
            type=DBType.MONGO.value,
            cluster__ex_params__database_pwd__isnull=False
        ).order_by("-created_date")

class IntegratedServicesList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_IntegratedServiceSerializer
    def get_queryset(self):
        return IntegratedService.objects.all()


    def list(self, request, *args, **kwargs):
        c123_node_objects = self.serializer_class(self.get_queryset(), many=True)
        if c123_node_objects:
            return Response(c123_node_objects.data)
        else:
            return Response({'Error':'There is no available integrated services'}, status=status.HTTP_404_NOT_FOUND)


class ClusterAvailableTree(generics.ListAPIView):
    permission_classes = [IsAuthenticated]
    serializer_class = c123_ClusterStatusSerializer

    def get_queryset(self):
        user = self.request.user
        queryset = c123_cluster.objects.annotate(
            has_nodes=Exists(
                c123_node.objects.filter(cluster=OuterRef('pk'), ind_deleted=False)
            )
        ).filter(has_nodes=True) \
          .select_related("created_by", "last_update_by") \
          .prefetch_related("c123_node_set__host")  # Preload related nodes & their hosts

        if not (user.is_superuser or user.groups.filter(name='Administrators').exists()):
            queryset = queryset.annotate(
                user_has_nodes=Exists(
                    c123_node.objects.filter(cluster=OuterRef('pk'), created_by=user, ind_deleted=False)
                )
            ).filter(user_has_nodes=True)

        return queryset

"""
class NodeListByDBType(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_NodeSerializer
    def get_queryset(self):
        db_type = self.kwargs['db_type']
        return c123_node.objects.filter(type = db_type)
    def list(self, request, *args, **kwargs):
        c123_filtered_node_objects = self.serializer_class(self.get_queryset(), many=True)
        if c123_filtered_node_objects:
            return Response(c123_filtered_node_objects.data)
        else:
            return Response({'Error':'There is no object'}, status=status.HTTP_404_NOT_FOUND)
class NodeListBySeverity(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_NodeSerializer
    def get_queryset(self):
        node_severity = self.kwargs['severity']
        return c123_node.objects.filter(severity = node_severity)
    def list(self, request, *args, **kwargs):
        c123_filtered_node_objects = self.serializer_class(self.get_queryset(), many=True)
        if c123_filtered_node_objects:
            return Response(c123_filtered_node_objects.data)
        else:
            return Response({'Error':'There is no object'}, status=status.HTTP_404_NOT_FOUND)
 """
class NodeSearchView(generics.ListAPIView):
    permission_classes = [IsAuthenticated]
    serializer_class = c123_NodeSerializer
    filter_backends = (filters.SearchFilter,)
    search_fields = ['host__name','name','port', 'type','cluster_role','created_by__username','status','severity']
    def get_queryset(self):
        user = self.request.user
        if user.is_superuser or user.groups.filter(name='Administrators').exists():
            return c123_node.objects.filter(org_id = user.profile.org_id)
        return c123_node.objects.filter(created_by = user,  org_id = user.profile.org_id)
class NodeListByTimeRange(generics.ListAPIView):
    permission_classes = [IsAuthenticated]
    serializer_class = c123_NodeSerializer
    def get_queryset(self):
        time_range = self.kwargs['time_range']
        user = self.request.user
        if user.is_superuser or user.groups.filter(name='Administrators').exists():
            nodes =  c123_node.objects.filter(org_id = user.profile.org_id)
        else:
            nodes = c123_node.objects.filter(created_by = user)
        if time_range == 5:
            five_minutes_ago = timezone.now() - timezone.timedelta(minutes=5)
            return nodes.filter(created_date__gte=five_minutes_ago)
        elif time_range == 1:
            one_hour_ago = timezone.now() - timezone.timedelta(hours=1)
            return c123_node.objects.filter(created_date__gte=one_hour_ago)
        elif time_range == 6:
            six_hours_ago = timezone.now() - timezone.timedelta(hours=6)
            return c123_node.objects.filter(created_date__gte=six_hours_ago)
        else:
            return Response({'Error':'Time range did not indicated'},status=status.HTTP_400_BAD_REQUEST)
class TaskLogList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class_task_log = c123_TaskLogSerializer
    def get_queryset(self):
        from urllib.parse import unquote
        node_id = self.kwargs.get('node_id')
        time_period = unquote(self.kwargs['time_period'])
        user = self.request.user
        now = timezone.now()
        match time_period:

            case Schedule.MINUTES_10.value:
                if user.is_superuser or user.groups.filter(name='Administrators').exists():
                    if node_id:
                        return TaskLog.objects.filter(target_id = node_id, created_date__gte = now - timezone.timedelta(seconds=600)).order_by("-task_id", "-created_date").distinct("task_id")
                    else:
                        return TaskLog.objects.filter(created_date__gte = now - timezone.timedelta(seconds=600)).order_by("-task_id", "-created_date").distinct("task_id")
                else:
                    if node_id:
                        return TaskLog.objects.filter(target_id = node_id, created_date__lte = now - timezone.timedelta(seconds=600), created_by = user).order_by("-task_id", "-created_date").distinct("task_id")
                    else:
                        return TaskLog.objects.filter(created_date__lte = now - timezone.timedelta(seconds=600), created_by = user).order_by("-task_id", "-created_date").distinct("task_id")

            case Schedule.HOUR.value:
                if user.is_superuser or user.groups.filter(name='Administrators').exists():
                    if node_id:
                        return TaskLog.objects.filter(target_id = node_id, created_date__gte = now - timezone.timedelta(seconds=3600)).order_by("-task_id", "-created_date").distinct("task_id")
                    else:
                        return TaskLog.objects.filter(created_date__gte = now - timezone.timedelta(seconds=3600)).order_by("-task_id", "-created_date").distinct("task_id")
                else:
                    if node_id:
                        return TaskLog.objects.filter(target_id = node_id, created_date__lte = now - timezone.timedelta(seconds=3600), created_by = user).order_by("-task_id", "-created_date").distinct("task_id")
                    else:
                        return TaskLog.objects.filter(created_date__lte = now - timezone.timedelta(seconds=3600), created_by = user).order_by("-task_id", "-created_date").distinct("task_id")

            case Schedule.HOURS_6.value:
                if user.is_superuser or user.groups.filter(name='Administrators').exists():
                    if node_id:
                        return TaskLog.objects.filter(target_id = node_id, created_date__gte = now - timezone.timedelta(seconds=21600)).order_by("-task_id", "-created_date").distinct("task_id")
                    else:
                        return TaskLog.objects.filter(created_date__gte = now - timezone.timedelta(seconds=21600)).order_by("-task_id", "-created_date").distinct("task_id")
                else:
                    if node_id:
                        return TaskLog.objects.filter(target_id = node_id, created_date__lte = now - timezone.timedelta(seconds=21600), created_by = user).order_by("-task_id", "-created_date").distinct("task_id")
                    else:
                        return TaskLog.objects.filter(created_date__lte = now - timezone.timedelta(seconds=21600), created_by = user).order_by("-task_id", "-created_date").distinct("task_id")
            case Schedule.DAY.value:
                if user.is_superuser or user.groups.filter(name='Administrators').exists():
                    if node_id:
                        return TaskLog.objects.filter(target_id = node_id, created_date__gte = now - timezone.timedelta(seconds=86400)).order_by("-task_id", "-created_date").distinct("task_id")
                    else:
                        return TaskLog.objects.filter(created_date__gte = now - timezone.timedelta(seconds=86400)).order_by("-task_id", "-created_date").distinct("task_id")
                else:
                    if node_id:
                        return TaskLog.objects.filter(target_id = node_id, created_date__lte = now - timezone.timedelta(seconds=86400), created_by = user).order_by("-task_id", "-created_date").distinct("task_id")
                    else:
                        return TaskLog.objects.filter(created_date__lte = now - timezone.timedelta(seconds=86400), created_by = user).order_by("-task_id", "-created_date").distinct("task_id")
            case _:
                if user.is_superuser or user.groups.filter(name='Administrators').exists():
                    if node_id:
                        return TaskLog.objects.filter(target_id = node_id, created_date__gte = now - timezone.timedelta(seconds=604800)).order_by("-task_id", "-created_date").distinct("task_id")
                    else:
                        return TaskLog.objects.filter(created_date__gte = now - timezone.timedelta(seconds=604800)).order_by("-task_id", "-created_date").distinct("task_id")
                else:
                    if node_id:
                        return TaskLog.objects.filter(target_id = node_id, created_date__lte = now - timezone.timedelta(seconds=604800), created_by = user).order_by("-task_id", "-created_date").distinct("task_id")
                    else:
                        return TaskLog.objects.filter(created_date__lte = now - timezone.timedelta(seconds=604800), created_by = user).order_by("-task_id", "-created_date").distinct("task_id")



    def list(self, request, *args, **kwargs):
        c123_task_log_objects = self.serializer_class_task_log(self.get_queryset(), many=True)
        if c123_task_log_objects:
            return Response(c123_task_log_objects.data)
        else:
            return Response({'Error':'There is no object'}, status=status.HTTP_404_NOT_FOUND)
class TaskLogTextList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class_task_log = c123_TaskLogTextSerializer
    def get_queryset(self):
        user = self.request.user
        if user.is_superuser or user.groups.filter(name='Administrators').exists():
            return TaskLog.objects.all()
        return TaskLog.objects.filter(created_by = user)

    def list(self, request, *args, **kwargs):
        c123_task_log_objects = self.serializer_class_task_log(self.get_queryset(), many=True)
        if c123_task_log_objects:
            return Response(c123_task_log_objects.data)
        else:
            return Response({'Error':'There is no object'}, status=status.HTTP_404_NOT_FOUND)
class TaskSearchView(generics.ListAPIView):
    permission_classes = [IsAuthenticated]
    serializer_class = c123_TaskLogSerializer
    filter_backends = (filters.SearchFilter,)
    search_fields = ['retcode', 'task_log_id', 'status', 'task','status', 'duration']
    def get_queryset(self):
        user = self.request.user
        if user.is_superuser or user.groups.filter(name='Administrators').exists():
            return Task.objects.filter(org_id = user.profile.org_id)
        return Task.objects.filter(created_by = user, org_id = user.profile.org_id)

class TaskStepLogList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class_task_step_log = c123_TaskStepLogSerializer
    def get_queryset(self):
        user = self.request.user
        if user.is_superuser or user.groups.filter(name='Administrators').exists():
            return TaskStepLog.objects.filter(org_id = user.profile.org_id).order_by('created_date')
        return TaskStepLog.objects.filter(created_by = user, org_id = user.profile.org_id).order_by('created_date')

    def list(self, request, *args, **kwargs):
        c123_task_step_log_objects = self.serializer_class_task_step_log(self.get_queryset(), many=True)
        if c123_task_step_log_objects:
            return Response(c123_task_step_log_objects.data)
        else:
            return Response({'Error':'There is no object'}, status=status.HTTP_404_NOT_FOUND)
class TaskStepLogDetail(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_TaskStepLogDetailSerializer

    def get_queryset(self):
        task_log_id = self.kwargs['task_log_id']
        return TaskStepLog.objects.filter(task_log_id=task_log_id).only('status', 'created_date', 'task', 'step', 'duration', 'retcode', 'retmsg', 'retoutput')
class HostMonitorCpuList(APIView):
    permission_classes = [IsAuthenticated]
    def get(self, request, format=None):
        last_6_hours = timezone.now() - timezone.timedelta(hours=6)
        host_monitor_obj = c123_host_monitor.objects.filter(metric_name="cpuusage", created_date__gte=last_6_hours).order_by("-created_date")
        host_monitor_obj_serializer = c123_HostMonitorSerializer(host_monitor_obj, many=True)
        if host_monitor_obj_serializer.data:
            return Response(host_monitor_obj_serializer.data)
        return Response({'Error':'There is no object'}, status=status.HTTP_204_NO_CONTENT)
class HostMonitorMemList(APIView):
    permission_classes = [IsAuthenticated]
    def get(self, request, format=None):
        last_6_hours = timezone.now() - timezone.timedelta(hours=6)
        host_monitor_obj = c123_host_monitor.objects.filter(metric_name="memusage",created_date__gt=last_6_hours).order_by("-created_date")
        host_monitor_obj_serializer = c123_HostMonitorSerializer(host_monitor_obj, many=True)
        if host_monitor_obj_serializer.data:
            return Response(host_monitor_obj_serializer.data)
        return Response({'Error':'There is no object'}, status=status.HTTP_204_NO_CONTENT)
class MemTotalLast(generics.RetrieveAPIView):
    permission_classes = [IsAuthenticated]
    queryset = c123_node.objects.all()
    def retrieve(self, request, pk):
        host = self.get_object().host
        mem_max = c123_HostMonitorSerializer(c123_host_monitor.objects.filter(host = host, metric_name = "memtotal").order_by("created_date").last())
        return JsonResponse({"mem_max": mem_max.data})
class NodeCreateView(generics.CreateAPIView):
    permission_classes =[IsAuthenticated]
    queryset = c123_node.objects.all()
    serializer_class = c123_AllNodeSerializer

class NodesBackupsList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_inventory_serializer
    def get_queryset(self):
       user = self.request.user
       if user.is_superuser or user.groups.filter(name='Administrators').exists():
           return c123_inventory.objects.filter(type = Inventory.BACKUP.value,  database__isnull = True)
       return c123_inventory.objects.filter(type = Inventory.BACKUP.value, created_by = user,  database__isnull = True)

class BackupsListByType(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_inventory_serializer
    def get_queryset(self):
       user = self.request.user
       type = self.kwargs['type']
       if user.is_superuser or user.groups.filter(name='Administrators').exists():
           if DBType.ORACLE_CDB.value in type:
               return  c123_inventory.objects.filter(type = Inventory.BACKUP.value, ex_params__belong_to__in=[DBType.ORACLE_CDB.value, DBType.ORACLE_PDB.value])
           else:
               return  c123_inventory.objects.filter(type = Inventory.BACKUP.value, ex_params__belong_to=type)
       if DBType.ORACLE_CDB.value in type:
           return c123_inventory.objects.filter(type = Inventory.BACKUP.value, ex_params__belong_to__in=[DBType.ORACLE_CDB.value, DBType.ORACLE_PDB.value], created_by = user)
       else:
           return c123_inventory.objects.filter(type = Inventory.BACKUP.value, ex_params__belong_to=type, created_by = user)

class NodesExportsList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_inventory_serializer
    def get_queryset(self):
       user = self.request.user
       if user.is_superuser or user.groups.filter(name='Administrators').exists():
           return  c123_inventory.objects.filter(type = Inventory.EXPORT.value, database__isnull = True)

       return c123_inventory.objects.filter(type = Inventory.EXPORT.value, created_by = user, database__isnull = True)

class ExportsListByType(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_inventory_serializer
    def get_queryset(self):
       user = self.request.user
       type = self.kwargs['type']
       if user.is_superuser or user.groups.filter(name='Administrators').exists():
           if DBType.ORACLE_CDB.value in type:
               return  c123_inventory.objects.filter(type = Inventory.EXPORT.value, ex_params__belong_to__in=[DBType.ORACLE_CDB.value, DBType.ORACLE_PDB.value])
           else:
               return  c123_inventory.objects.filter(type = Inventory.EXPORT.value, ex_params__belong_to=type)
       if DBType.ORACLE_CDB.value in type:
           return c123_inventory.objects.filter(type = Inventory.EXPORT.value, ex_params__belong_to__in=[DBType.ORACLE_CDB.value, DBType.ORACLE_PDB.value], created_by = user)
       else:
           return c123_inventory.objects.filter(type = Inventory.EXPORT.value, ex_params__belong_to=type, created_by = user)



class DatabasesBackupsList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_inventory_serializer
    def get_queryset(self):
        user = self.request.user
        if user.is_superuser or user.groups.filter(name='Administrators').exists():
             return c123_inventory.objects.filter(type = Inventory.BACKUP.value, database__isnull = False)
        return c123_inventory.objects.filter(type = Inventory.BACKUP.value, created_by = user,  database__isnull = False)

class DatabasesExportsList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_inventory_serializer
    def get_queryset(self):
       user = self.request.user
       if user.is_superuser or user.groups.filter(name='Administrators').exists():
           return c123_inventory.objects.filter(type = Inventory.EXPORT.value, database__isnull = False)
       return c123_inventory.objects.filter(type = Inventory.EXPORT.value, created_by = user,  database__isnull = False)

@api_view(['POST'])
@permission_classes([IsAuthenticated])
def list_node_logs(request):
    try:
        node_id = request.data.get('node_id')
        if not node_id:
            raise APIException400("node_id is required")

        node = c123_node.objects.filter(pk=node_id, ind_deleted=False).first()
        if not node:
            raise APIException410(f"node id {node_id} not available")

        host = node.host
        ssh = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
        if isinstance(ssh, tuple):
            if hasattr(ssh[1], 'strerror'):
                raise APIException401(ssh[1].strerror)
            raise APIException401(ssh[1])

        # Modify the find command to include modification date
        find_command = f"""
        sudo find {node.dir_123Cluster}/log -type f -name "*.log" -exec stat --format '%n %s %Y' {{}} \; |
        awk 'BEGIN {{print "["}}
        {{
            size=$2;
            filename=substr($1, match($1, /[^\/]*$/));
            mod_time=strftime("%Y-%m-%d %H:%M:%S", $3);

            if (size < 1024) size_str = size "B";
            else if (size < 1048576) size_str = int(size / 1024) "KB";
            else if (size < 1073741824) size_str = int(size / 1048576) "MB";
            else size_str = int(size / 1073741824) "GB";

            print (NR>1 ? "," : "") "{{\\"path\\": \\"" $1 "\\", \\"filename\\": \\"" filename "\\", \\"size\\": \\"" size_str "\\", \\"modified_date\\": \\"" mod_time "\\"}}"
        }}
        END {{print "]"}}'
        """

        jsn = json.loads(remote_command(ssh, find_command, True), strict=False)
        if jsn['stderr'] != '':
            raise APIException403(jsn['stderr'])
        if not jsn['stdout'].strip():
            node_log_files = []
        else:
            node_log_files = json.loads(jsn['stdout'])

        expires = int(time.time()) + 3600  # 1 hour expiry
        secret = settings.SECRET_KEY.encode()
        base_url = request.build_absolute_uri(reverse('download_node_log'))

        for f in node_log_files:
            file_path = f.get('path')
            signature = base64.urlsafe_b64encode(
                hmac.new(secret, f"{file_path}{expires}".encode(), hashlib.sha256).digest()
            ).decode()
            f["download_url"] = f"{base_url}?{urlencode({'file_path': file_path, 'node_id': node_id, 'expires': expires, 'signature': signature})}"

        return JsonResponse({"node_log_files": node_log_files}, status=200)

    except Exception as e:
        raise APIException500(str(e))


class DownloadNodeLogs(APIView):
    authentication_classes = []  # Disable authentication
    permission_classes = [AllowAny]  # Allow any user to access this endpoint

    def get(self, request, *args, **kwargs):
        try:
            # Extract URL parameters
            file_path = request.GET.get('file_path')
            expires = request.GET.get('expires')
            signature = request.GET.get('signature')

            # Validate parameters
            if not file_path or not expires or not signature:
                return JsonResponse({"error": "Missing parameters"}, status=400)

            # Validate expiration
            if int(expires) < int(time.time()):
                return JsonResponse({"error": "Link expired"}, status=403)

            # Validate signature
            secret = settings.SECRET_KEY.encode()
            expected_signature = base64.urlsafe_b64encode(
                hmac.new(secret, f"{file_path}{expires}".encode(), hashlib.sha256).digest()
            ).decode()

            if signature != expected_signature:
                return JsonResponse({"error": "Invalid signature"}, status=403)

            # Open the SSH connection
            node_id = request.GET.get("node_id")
            node = c123_node.objects.filter(pk=node_id, ind_deleted=False).first()
            if not node:
                return JsonResponse({"error": f"Node id {node_id} not available"}, status=410)

            host = node.host
            ssh = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
            if isinstance(ssh, tuple):
                return JsonResponse({"error": "SSH Connection Failed"}, status=401)

            # Open the SFTP connection
            sftp = ssh.open_sftp()
            try:
                remote_file = sftp.open(file_path, 'rb')
            except Exception as e:
                return JsonResponse({"error": f"Failed to open file: {str(e)}"}, status=500)

            file_size = sftp.stat(file_path).st_size

            # File streaming in chunks
            def file_iterator(file_obj, chunk_size=8192):
                for chunk in iter(lambda: file_obj.read(chunk_size), b''):
                    yield chunk

            # Create the response for downloading the file
            response = StreamingHttpResponse(file_iterator(remote_file), content_type='application/octet-stream')
            response['Content-Disposition'] = f'attachment; filename="{os.path.basename(file_path)}"'
            response['Content-Length'] = str(file_size)

            return response

        except Exception as e:
            return JsonResponse({"error": str(e)}, status=500)

@api_view(['PUT'])
@permission_classes((IsAuthenticated, ))
def cluster_update(request):
    db_type = request.data['db_type']
    serializer = None
    if(DBType.ORACLE_CDB.value in db_type):
        cluster_sizing, created = OracleClusterSizing.objects.get_or_create(db_type = DBType.ORACLE_CDB.value)
        serializer = OracleClusterSizingSerializer(cluster_sizing, data=request.data)
        if(serializer.is_valid()):
            serializer.save()
            return Response(status=status.HTTP_200_OK)
        else:
            return Response(data=serializer.errors, status=status.HTTP_400_BAD_REQUEST)
    elif (DBType.MYSQL.value in db_type):
         cluster_sizing, created = ClusterSizing.objects.get_or_create(db_type = DBType.MYSQL.value)
         serializer = ClusterSizingSerializer(cluster_sizing,  data=request.data)
         if(serializer.is_valid()):
            serializer.save()
            return Response(status=status.HTTP_200_OK)
         else:
            return Response(data=serializer.errors, status=status.HTTP_400_BAD_REQUEST)
    elif (DBType.MONGO.value in db_type):
        cluster_sizing, created = ClusterSizing.objects.get_or_create(db_type = DBType.MONGO.value)
        serializer = ClusterSizingSerializer(cluster_sizing, data=request.data)
        if(serializer.is_valid()):
            serializer.save()
            return Response(status=status.HTTP_200_OK)
        else:
            return Response(data=serializer.errors, status=status.HTTP_400_BAD_REQUEST)
    elif (DBType.POSTGRES.value in db_type):
        cluster_sizing, created = ClusterSizing.objects.get_or_create(db_type = DBType.POSTGRES.value)
        serializer = ClusterSizingSerializer(cluster_sizing, data=request.data)
        if(serializer.is_valid()):
            serializer.save()
            return Response(status=status.HTTP_200_OK)
        else:
            return Response(data=serializer.errors, status=status.HTTP_400_BAD_REQUEST)
    elif (DBType.MARIADB.value in db_type):
        cluster_sizing, created = ClusterSizing.objects.get_or_create(db_type = DBType.MARIADB.value)
        serializer = ClusterSizingSerializer(cluster_sizing, data=request.data)
        if(serializer.is_valid()):
            serializer.save()
            return Response(status=status.HTTP_200_OK)
        else:
            return Response(data=serializer.errors, status=status.HTTP_400_BAD_REQUEST)
    elif (DBType.ELASTIC.value in db_type):
        cluster_sizing, created = ClusterSizing.objects.get_or_create(db_type = DBType.ELASTIC.value)
        serializer = ClusterSizingSerializer(cluster_sizing, data=request.data)
        if(serializer.is_valid()):
            serializer.save()
            return Response(status=status.HTTP_200_OK)
        else:
            return Response(data=serializer.errors, status=status.HTTP_400_BAD_REQUEST)
    elif (DBType.REDIS.value in db_type):
        cluster_sizing, created = ClusterSizing.objects.get_or_create(db_type = DBType.REDIS.value)
        serializer = ClusterSizingSerializer(cluster_sizing, data=request.data)
        if(serializer.is_valid()):
            serializer.save()
            return Response(status=status.HTTP_200_OK)
    elif (DBType.MSSQL.value in db_type):
        cluster_sizing, created = ClusterSizing.objects.get_or_create(db_type = DBType.MSSQL.value)
        serializer = ClusterSizingSerializer(cluster_sizing, data=request.data)
        if(serializer.is_valid()):
            serializer.save()
            return Response(status=status.HTTP_200_OK)
        else:
            return Response(data=serializer.errors, status=status.HTTP_400_BAD_REQUEST)
    return Response(status=status.HTTP_304_NOT_MODIFIED)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def checkDiskExists(request):
    """check if a directory exists for a node export operation"""
    node_id = request.data.get('node_id')
    obj_type = request.data.get("obj_type")
    path = request.data.get('path')[1] if obj_type == ClusterRoles.CDB.value or obj_type == ClusterRoles.PDB.value else request.data.get('path')
    if ObjectType.DATABASE.value in obj_type:
        db = c123_database.objects.filter(pk = node_id).first()
        if not db:
            raise APIException410(f"database id {node_id} not available")
        node = db.node
    else:
        node = c123_node.objects.filter(pk=node_id, ind_deleted = False).first()
        if not node:
            raise APIException410(f"node id {node_id} not available")

        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")

        # host = app_host if node.type in [DBType.MSSQL.value, DBType.POSTGRES.value] else node.host

        # client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
        # if isinstance(client, tuple):
        #     if hasattr(client[1], 'strerror'):
        #         raise APIException401(client[1].strerror)
        #     raise APIException401(client[1])

        if path.startswith('~'):
            tilda_path = find_tilda(node.host)
            path = path.replace('~', tilda_path)

            if "cannot stat" in jsn["stderr"].lower():
                jsn = json.loads(local_command(f"sudo mkdir -p '{path}'", True), strict=False)
                error = str(jsn['stderr'])
                if error != '':
                    raise APIException400(error)

    return JsonResponse({}, status = 200)


@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def IsNodeHostReachable(request):
    node = c123_node.objects.filter(pk = request.data.get('node_id')).first()
    return JsonResponse({'reachable': not isinstance(check_host_connection(node), tuple) if node else False}, status = 200)


@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckCpuRamCurrent(request):

    node_id = request.data.get('node_id')
    node = c123_node.objects.filter(pk = node_id, ind_deleted = False).first()
    if not node:
        raise APIException410(f"node {node_id} is not available")

    host = node.host
    node_type = node.type

    if node_type not in [DBType.MSSQL.value, DBType.ELASTIC.value, DBType.REDIS.value]:

        client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
        if isinstance(client, tuple):
            error = client[1].strerror if hasattr(client[1], "strerror") else str(client[1])
            raise APIException401(error)


        if node_type != DBType.POSTGRES:
            jsn = json.loads(remote_command(client, f"""sudo netstat -tunlp | grep '{node.port}' | awk '{{print $7}}' | awk -F'/' '{{print $1}}'  | sort -u | xargs -I{{}} bash -c 'awk "/^VmRSS:/ {{print \$2}}" /proc/{{}}/status && awk "{{print (\$14+\$15)/100}}" /proc/{{}}/stat' | awk '{{total_mem+=$1; total_cpu+=$2}} END {{print total_cpu, total_mem}}'""", True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if jsn['stderr'] != '':
                return JsonResponse(jsn, status = 404)

            measurements = jsn['stdout'].split(' ')
            if len(measurements) > 1:
                return JsonResponse({'cpu': measurements[0], 'ram': measurements[1]}, status = 200)
            else:
                return JsonResponse({'cpu': 'N/A', 'ram': 'N/A'}, status = 204)

        else:
            command = "sudo ps -C postgres -o rss= | awk '{sum+=$1} END {print sum}'"

            jsn = json.loads(remote_command(client, command, True), strict=False)
            if(jsn['stderr'] != ''):
                return JsonResponse(jsn, status=404)

            ram = format_size(jsn['stdout'])

            command = r"""total_usage=$(top -b -n 1 | grep 'postgres' | awk '{sum += $9} END {print sum}')
        core_count=$(nproc)
        awk -v usage="$total_usage" -v cores="$core_count" 'BEGIN {printf "%i\n", usage / cores}'"""
            jsn = json.loads(remote_command(client, command, True), strict=False)
            if(jsn['stderr'] != ''):
                return JsonResponse(jsn, status=404)
            cpu = jsn['stdout']
            return JsonResponse({'cpu': cpu, 'ram': ram}, status = 200)
    elif node_type == DBType.REDIS.value:
        client = connect_to_redis(node)
        if isinstance(client, tuple):
            error = client[1].strerror if hasattr(client, 'strerror') else str(client[1])
            raise APIException401(error)

        try:
            info = client.info()
            ram = format_size(float(info['used_memory_rss']) / 1024.0)
            cpu = round(float(info['used_cpu_sys']) + float(info['used_cpu_user']) + float(info['used_cpu_sys_children']) + float(info['used_cpu_user_children']))
            return JsonResponse({'cpu': cpu, 'ram': ram}, status = 200)
        except Exception as ex:
            error = ex.strerror if hasattr(ex, 'strerror') else str(ex)
            raise APIException403(error)

    elif node_type == DBType.ELASTIC.value:
        client = create_elasticsearch_insecure_connection(node)
        if isinstance(client, tuple):
            if hasattr(client[1], 'strerror'):
                error = client[1].strerror
            else:
                error = str(client[1])
            raise APIException401(error)

        try:
            node_data = client.cat.nodes(h="http,id,v,u,du,rc,cpu", format="json", full_id=True, bytes='kb')
            current_node =  [i for i in node_data if node.name in i["http"]]
            if len(current_node) > 0:
                ram = format_size(float(current_node["rc"]) / 1024)
                cpu = current_node["cpu"]
                return JsonResponse({'cpu': cpu, 'ram': ram}, status = 200)

            else:
                return JsonResponse({'cpu': 'N/A', 'ram': 'N/A'}, status = 204)
        except Exception as ex:
            error = ex.strerror if hasattr(ex, 'strerror') else str(ex)
            raise APIException403(error)

    elif node_type == DBType.MSSQL.value:
        return JsonResponse({'cpu': 'N/A', 'ram': 'N/A'}, status = 204)



@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def MarkNodeDeleted(request):
    node_id = request.data.get('node_id')
    return JsonResponse({"nodes_updated" : c123_node.objects.filter(pk = node_id).update(ind_deleted = True), "databases_updated": c123_database.objects.filter(node__id = node_id).update(ind_deleted = True)},  status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def IsRunningCompleted(request):
    tasks = Task.objects.filter(status = TaskState.ON_DEMAND.value, task_id__in =  [t['task_id'] for t in request.data])
    tasks_exists = tasks.exists()
    if tasks_exists:
        live_replica = tasks.filter(displayment__icontains='live replica', created_by = request.user)
        if live_replica.exists():
            subject = "Ongoing live replica notification"
            replica_notice = '''
            Dear {{user}}

            We would like to inform you that the {{displayment}} is now established.
            Thank you for your attention to this matter.
            Best regards,
            123Cluster
            '''
            template = Template(replica_notice)


            for t in live_replica:
                user = t.created_by
                context = Context({"user":  user.username, "displayment":t.displayment})
                message = template.render(context)
                send_mail(subject, message, [user.email])

    return JsonResponse({"refresh" : tasks_exists},  status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def checkInventoryExists(request):
    return JsonResponse({'exists': c123_inventory.objects.filter(node__name = request.data.get("name"), ex_params__belong_to = request.data.get('type')).exists()},  status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def isQuotaExceed(request):
    """check if user exceed quota"""
    quota = request.data.get("quota")
    obj_type = request.data.get("obj_type")
    user = request.user
    user_nodes_by_type = c123_node.objects.filter(type__contains=obj_type, ind_deleted = False, created_by = user).count()
    exceed = False if user_nodes_by_type < quota else True
    return JsonResponse({"exceed": exceed}, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def checkExportFile(request):
    """check if a directory exists for a node export operation"""
    node_id = request.data.get('node_id')
    obj_type = request.data.get('obj_type')
    schema = request.data.get('schema')
    pdb_name = request.data.get('pdb_name')
    path = request.data.get('path')[1] if obj_type == ClusterRoles.CDB.value or obj_type == ClusterRoles.PDB.value else request.data.get('path')
    if ObjectType.DATABASE.value in obj_type:
        db = c123_database.objects.filter(pk = node_id).first()
        if not db:
            raise APIException410(f"database id {node_id} not available")
        node = db.node
    else:
         node = c123_node.objects.filter(pk=node_id, ind_deleted = False).first()
         if not node:
            raise APIException410(f"node id {node_id} not available")
    name_to_use =  pdb_name if obj_type == ClusterRoles.PDB.value else db.name if ObjectType.DATABASE.value in obj_type else node.name
    # file name variables
    now = timezone.now().strftime('%d%m%Y%H%M')
    file_name_list = list(chain(schema, [name_to_use], [now]))  if obj_type == ClusterRoles.CDB.value or obj_type == ClusterRoles.PDB.value else  list(chain([name_to_use], [now]))
    file_name = "_".join(file_name_list) + ".dmp"
    full_path = path + "/" + file_name
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    host = app_host if node.type in [DBType.MSSQL.value, DBType.POSTGRES_K8S.value] else node.host
    client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            raise APIException401(client[1].strerror)
        raise APIException401(client[1])
    jsn = json.loads(remote_command(client, f"sudo ls -ltr '{full_path}'", True), strict=False)
    # reverse logic here
    if(jsn['stderr'] != ''):
        # if the command completes with error it means the directory exists [which is good]
        return JsonResponse({'full_path': full_path, 'file_name': file_name}, status=200)
    else:
        # otherwise we created a new directory and the task failed
        # since it means we can't find the export file
        return JsonResponse({'detail': jsn['stdout']}, status=404)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def checkDiskSpace(request):
    """check if a directory exists for a node export operation"""
    node_id = request.data.get('node_id')
    required_space = request.data.get('required_space')
    obj_type = request.data.get("obj_type")
    path = request.data.get('path')[1] if obj_type == ClusterRoles.CDB.value or obj_type == ClusterRoles.PDB.value else request.data.get('path')
    if ObjectType.DATABASE.value in obj_type:
        db = c123_database.objects.filter(pk = node_id).first()
        if not db:
            raise APIException410(f"database id {node_id} not available")
        node = db.node

    else:
         node = c123_node.objects.filter(pk=node_id, ind_deleted = False).first()
         if not node:
            raise APIException410(f"node id {node_id} not available")

    response = check_space(node = node, path = path, required_space = required_space)
    return JsonResponse(response, status = response['status'])

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def checkFreeIp(request):
    """check if ip is not occupied on proxmox host"""
    service_name = request.data.get('service_name')
    service = IntegratedService.objects.filter(name = service_name, type = IntegratedServiceTypes.PROXMOX.name).first()
    if not service:
        raise APIException404(f"service {service_name} does not exist")
    ex_params = service.ex_params
    path = os.path.join(os.getcwd(), 'list_ct.py')
    result = run_python_script(path, [service.host_name, ex_params.get('proxmox_user'), ex_params.get('proxmox_token_name'), ex_params.get('proxmox_token_value'), str(ex_params.get('verify_ssl'))])
    if result['excode'] != 0:
        raise APIException500(f"exit code: {result['excode']} error: {result['stdout']}")

    import re
    used_ips = [ip.group() for s in result['stdout'].splitlines() if (ip := re.search(r'\b(?:[0-9]{1,3}\.){3}[0-9]{1,3}\b', s))]
    print(used_ips)
    available_ip = get_random_available_ip('172.16.0.0/24', used_ips)
    if not available_ip:
        raise APIException409('No available ips present')

    return JsonResponse({'available_ip': available_ip}, status = 200)


class CreateDemoUser(APIView):
    authentication_classes = []  # Disable authentication
    permission_classes = [AllowAny]  # Allow any user to access this endpoint

    def get(self, request, *args, **kwargs):
        try:
            email = request.GET.get('email').strip()
            valid_email = is_valid_email(email)
            if not valid_email:
                raise APIException400(f"email {email} is invalid")
            user_exists = User.objects.filter(email = email, is_active = True).exists()
            if user_exists:
                raise APIException406(f"Email {email} has been already taken!")
            username = generate_human_readable_username()
            password = generate_random_password()
            user = User.objects.create_user(
                username = username,
                email = email,
                password = password,
                is_active = True,
                is_superuser = False)
            current_site = get_current_site(request)
            profile,created = Profile.objects.get_or_create(user = user, org_id = Organization.objects.first(), is_demo = True)
            privileges, created = UserPrivileges.objects.get_or_create(user = user)
            notifications, created = UserNotifications.objects.get_or_create(user = user, email = email)
            return JsonResponse({'username': username, 'password': password}, status = status.HTTP_201_CREATED)

        except Exception as e:
            raise APIException500(str(e))
        
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def signup(request):
    user_dict = request.data["user_dict"]
    user_dict["username"] =  user_dict["username"].lower()
    if User.objects.filter(username = user_dict["username"]).exists():
        return JsonResponse({'result': 'username already taken!'})


    try:
        notice_dict = request.data["notice_dict"]
        user = User.objects.create_user(
        user_dict["username"],  notice_dict["email"], user_dict["password"])
        user.is_active = False
        (profile,created) = Profile.objects.get_or_create(user = user, org_id = Organization.objects.first())
        user.save()
        account_activation_token = TokenGenerator()
    except Exception as e:
            error, = e.args
            messages.error(request, error)
            return Response(status=status.HTTP_400_BAD_REQUEST)

    try:
        priv_dict = request.data["priv_dict"]
        if 'True' in priv_dict["is_superuser"]:
            admins = Group.objects.get(name='Administrators')
            admins.user_set.add(user)
            user.is_superuser = True
            user.save()
        if('False' in priv_dict["is_superuser"]):
            del priv_dict["is_superuser"]
            user_privileges = UserPrivileges.objects.create(**request.data["priv_dict"])
            user_privileges.user = user
            user_privileges.save()

        if isinstance(notice_dict['email'], str) and notice_dict['email'] and not notice_dict['email'].isspace():
            if UserNotifications.objects.filter(email = notice_dict['email']).exists():
                    return JsonResponse({"result": "email already taken!"})
            user_notifications = UserNotifications.objects.create(**notice_dict)
            user_notifications.user = user
            user_notifications.save()
            current_site = get_current_site(request)
            mail_subject = 'Activation link has been sent to your email id'
            template = Template('''{% autoescape off %}
    Hi {{ user.username }},
    Please click on the link to confirm your registration,
    https://{{ domain }}{% url 'activate' uidb64=uid token=token %}
    {% endautoescape %} ''')
            context = Context({
                'user': user,
                'domain': current_site.domain,
                'uid':urlsafe_base64_encode(force_bytes(user.pk)),
                'token':account_activation_token.make_token(user),
            })
            message = template.render(context)

            send_mail(mail_subject, message, [user_notifications.email])
        else:
            user.is_active = True
            user.save()
    except Exception as e:
        error, = e.args
        messages.error(request, error)
        return Response(status=status.HTTP_206_PARTIAL_CONTENT)

    return Response(status=status.HTTP_201_CREATED)

def activate(request, uidb64, token):
    try:
        uid = force_str(urlsafe_base64_decode(uidb64))
        user = User.objects.filter(pk=uid).first()
        if user:
            account_activation_token = TokenGenerator()
            activated = account_activation_token.check_token(user, token)
            if activated:
                user.is_active = True
                user.save()
                return HttpResponse('Thank you for your email confirmation. Now you can login your account.')
            else:
                return HttpResponse(f'Invalid activation token: {token}')

        else:
            return HttpResponse(f"Invalid user id: {uid}")

    except Exception as e:
         return HttpResponse(str(e))

@api_view(['GET', 'PUT'])
@permission_classes((IsAuthenticated, ))
def user_update(request, pk):
    the_user = User.objects.get(pk=pk)
    current = the_user == request.user
    is_superuser = the_user.groups.filter(name='Administrators').exists() or the_user.is_superuser
    if request.method == 'PUT':
        user_dict = request.data["user_dict"]
        if user_dict.get('username'):
            user_dict["username"] =  user_dict["username"].lower()
        notice_dict = request.data["notice_dict"]
        del notice_dict["id"]
        if user_dict["password"] and (is_superuser or current):
            the_user.set_password(user_dict["password"])
            the_user.save()
            update_session_auth_hash(request, the_user)
        if not is_superuser and (request.user.is_superuser or request.user.groups.filter(name='Administrators').exists()):
            priv_dict = request.data["priv_dict"]
            if('False' in priv_dict["is_superuser"]):
                del priv_dict["is_superuser"]
                UserPrivileges.objects.filter(user=the_user).update(**priv_dict)
            elif 'True' in priv_dict["is_superuser"]:
                admins = Group.objects.get(name='Administrators')
                admins.user_set.add(the_user)
                UserPrivileges.objects.filter(user=the_user).delete()
                the_user.is_superuser = True
                the_user.is_staff = True
                the_user.save()

        user_notifications = UserNotifications.objects.filter(email = notice_dict["email"])
        if user_notifications :
            if UserNotifications.objects.filter(email = notice_dict["email"]).exclude(user = the_user).first() and notice_dict["email"] != "":
                return JsonResponse({"result": "email already taken!"})
            user_notifications.update(**notice_dict)
        else:
            user_notifications.create(user = the_user, email = notice_dict["email"], nodes_resolved_alerts = notice_dict["nodes_resolved_alerts"], completed_actions_alerts =  notice_dict["completed_actions_alerts"])
        return Response(status=status.HTTP_202_ACCEPTED)
    else:
        user_serializer = UserSerializer(the_user)
        user_notifications = UserNotifications.objects.filter(user = the_user).last()
        if not user_notifications:
            user_notifications = UserNotifications.objects.create(user = the_user, email = the_user.email, nodes_resolved_alerts = "ONU", completed_actions_alerts = "ONU")
        notice_dict_serializer = UserNotificationsSerializer(user_notifications)
        if(not UserPrivileges.objects.filter(user = the_user).exists() and is_superuser):
            return JsonResponse({"user_dict": user_serializer.data, "notice_dict":  notice_dict_serializer.data,  'is_superuser': is_superuser, "current": current})
        else:
            privileges, created = UserPrivileges.objects.get_or_create(user = the_user)
            priv_dict_serializer = UserPrivilegesSerializer(privileges)
            return JsonResponse({"user_dict": user_serializer.data, "notice_dict":  notice_dict_serializer.data, "priv_dict": priv_dict_serializer.data, 'is_superuser': is_superuser, "current": current})

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def VerifyBackup(request):
    """check if a backup folder has modified"""
    backup_path = request.data.get("backup_path")
    backup_date = request.data.get("backup_date")
    host = c123_host.objects.filter(ex_params__myself = True).first()
    if not host:
        raise APIException404(f"app host is not found")
    client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            raise APIException401(client[1].strerror)
        raise APIException401(client[1])
    jsn = json.loads(remote_command(client, f"sudo find '{backup_path}' -newermt '{backup_date}'", True), strict=False)

    if(jsn['stderr'] != ''):
        raise APIException404(jsn['stderr'])
    elif jsn["stdout"].strip() == '':
            return JsonResponse({"ok": False}, status=200)
    elif jsn["stdout"].strip() != '':
        return JsonResponse({"ok": True}, status=200)

@api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def GetDatabaseUsers(request, id):
    node = c123_node.objects.filter(pk = id, ind_deleted = False).first()
    if not node:
        raise APIException410(f"db {id} is not available")

    host = node.host
    db_users = []
    owner_description = "owner of this database, select any table or view in this database, select any table or view and update any table in this database"
    write_description = "dml user of this database, select any table or view in this database, insert/update/delete to any table in this database"
    read_description = "read user of this database, select any table or view in this database"
    try:
        match node.type:
            case DBType.POSTGRES.value:
                client = get_postgres_connection(node)
                if isinstance(client, tuple):
                    error = client[1].strerror if hasattr(client[1], 'strerror') else str(client[1])
                    raise APIException401(error)
                cursor = client.cursor(cursor_factory=RealDictCursor)
                query = f'''SELECT array_agg(rolname) AS database_users FROM pg_roles;'''

                cursor.execute(query)

                db_users_raw = cursor.fetchone()
                print(db_users_raw)
                if db_users_raw:
                    db_users_raw = db_users_raw.get('database_users')
                    for user in db_users_raw:
                        if 'owner_user' in user:
                            db_users.append({"name":  user, "description": owner_description})
                        elif 'write_user' in user:
                            db_users.append({"name":  user, "description": write_description})
                        elif 'read_user' in user:
                            db_users.append({"name":  user, "description": read_description})
                        else:
                            db_users.append({"name":  user, "description": ''})
            case DBType.MARIADB.value | DBType.MYSQL.value:
                client = create_mysql_connection(node)
                if isinstance(client, tuple):
                    raise APIException401(str(client[1]))

                cursor = client.cursor(dictionary= True)
                query = '''SELECT DISTINCT
    `GRANTEE`
FROM
    `INFORMATION_SCHEMA`.`SCHEMA_PRIVILEGES`
WHERE
    `GRANTEE` NOT LIKE '%%\_%%' ESCAPE '|'
    AND `GRANTEE` NOT LIKE 'mysql.%%'
    AND `GRANTEE` NOT LIKE 'root@%%'
    AND `GRANTEE` NOT LIKE 'anonymous%%'
    AND `GRANTEE` NOT LIKE 'pma_%%'
    AND `GRANTEE` NOT LIKE 'phpmyadmin%%'
    AND `GRANTEE` NOT LIKE 'debian-%%'
    AND `GRANTEE` NOT LIKE 'mysql.session@%%'
    AND `GRANTEE` NOT LIKE 'mysql.sys@%%';'''

                cursor.execute(query)
                db_users_raw = cursor.fetchall()
                for user in db_users_raw:
                    user = user['GRANTEE'].strip("\'")
                    if '_owner' in user:
                        db_users.append({"name":  user, "description": owner_description})
                    elif '_write' in user:
                        db_users.append({"name":  user, "description": write_description})
                    elif '_read' in user:
                        db_users.append({"name":  user, "description": read_description})
                    else:
                        db_users.append({"name":  user, "description": ''})
            case DBType.MONGO.value:
                client = create_mongo_connection(node)
                if isinstance(client, tuple):
                    if hasattr(client[1], 'strerror'):
                        raise APIException401(client[1].strerror)
                    raise APIException401(client[1])

                users = client.admin.command('usersInfo')
                for user in users['users']:
                    user = user['user'].strip()
                    if '_owner' in user:
                        db_users.append({"name":  user, "description": owner_description})
                    elif '_write' in user:
                        db_users.append({"name":  user, "description": write_description})
                    elif '_read' in user:
                        db_users.append({"name":  user, "description": read_description})
                    else:
                        db_users.append({"name":  user, "description": ''})
            case DBType.ELASTIC.value:
                client = create_elasticsearch_insecure_connection(node)
                if isinstance(client, tuple):
                    raise APIException401(str(client[1]))

                roles_info = client.security.get_role()
                for user, role_info in roles_info.items():
                    indices = role_info.get('indices')
                    if indices:
                        for pattern in indices:
                            names = pattern.get('names')
                            if names and len(names) > 0:
                                if '_owner' in user:
                                    db_users.append({"name":  user, "description": owner_description})
                                elif '_write' in user:
                                    db_users.append({"name":  user, "description": write_description})
                                elif '_read' in user:
                                    db_users.append({"name":  user, "description": read_description})
                                else:
                                    db_users.append({"name":  user, "description": ''})

            case DBType.MSSQL.value:
                 client = connect_to_sqlserver_instance(hostname=host.name, uid = node.database_user, pwd=node.database_pwd, prt=node.port)
                 if isinstance(client, tuple):
                    if hasattr(client[1], 'strerror'):
                            raise APIException401(client[1].strerror)
                    raise APIException401(client[1])

                 cursor = client.cursor()
                 cursor.execute("SELECT name FROM sys.database_principals WHERE type_desc = 'SQL_USER' AND NAME NOT IN('dbo', 'guest', 'INFORMATION_SCHEMA', 'sys')")
                 for row in cursor.fetchall():
                       if '_owner' in row.name:
                           db_users.append({"name":  row.name, "description": owner_description})
                       elif '_write' in row.name:
                           db_users.append({"name":  row.name, "description": write_description})
                       elif '_read' in row.name:
                           db_users.append({"name":  row.name, "description": read_description})
                       else:
                           db_users.append({"name":  row.name, "description": ''})

        for user in db_users:
            if node.type != DBType.MSSQL.value:
                match(node.type):
                    case DBType.MONGO.value:
                        db_name = "admin"
                    case DBType.POSTGRES.value:
                        db_name = "postgres"
                    case DBType.MARIADB.value | DBType.MYSQL.value:
                        db_name = "mysql"
                    case _ :
                        pass

                user['connection_string'] = generate_connection_string(node.type, host.name, node.port, user['name'], None if node.type == DBType.REDIS.value else db_name)
            else:
                user['connection_string'] = connect_to_sqlserver_instance(hostname=host.name, uid = user['name'], pwd=node.database_pwd, prt=node.port, db = 'master', dry_run = True)
    except Exception as e:
         raise APIException(detail = e.strerror if hasattr(e, 'strerror') else str(e), code = e.code if hasattr(e, 'code') else 500)

    return JsonResponse(db_users, safe = False, status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def EditDatabaseUser(request):
    node_id = request.data.get('node_id')
    user = request.data.get('user')
    new_password = request.data.get('new_password')
    if new_password:
        new_password = format_password(new_password)
    node = c123_node.objects.filter(pk = node_id, ind_deleted = False).first()
    if not node:
        raise APIException410(f"node {node_id} is not available")

    host = node.host
    try:
        match node.type:
            case DBType.POSTGRES.value:
                client = get_postgres_connection(node)
                if isinstance(client, tuple):
                    error = client[1].strerror if hasattr(client[1], 'strerror') else str(client[1])
                    raise APIException401(error)

                cursor = client.cursor()
                if new_password:
                    cursor.execute("ALTER USER %s WITH PASSWORD '%s'" % (user, new_password))
                    client.commit()

            case DBType.MYSQL.value | DBType.MARIADB.value:

                client = create_mysql_connection(node)
                if isinstance(client, tuple):
                    if hasattr(client[1], 'strerror'):
                        raise APIException401(client[1].strerror)
                    raise APIException401(client[1])

                cursor = client.cursor()
                if new_password:
                     query = "ALTER USER '%s' IDENTIFIED WITH mysql_native_password BY '%s'" if node.type == DBType.MYSQL.value else "SET PASSWORD FOR `%s` = PASSWORD('%s')"
                     cursor.execute(query % (user, new_password))
                     cursor.execute("FLUSH PRIVILEGES;")
                     client.commit()
            case DBType.MONGO.value:
                client = create_mongo_connection(node)
                if isinstance(client, tuple):
                    if hasattr(client[1], 'strerror'):
                        raise APIException401(client[1].strerror)
                    raise APIException401(client[1])

                user_data = client.admin.command('usersInfo', username = user)['users'][0]
                if new_password:
                    user_data['pwd'] = new_password
                    client.admin.command('updateUser', user, **user)

            case DBType.ELASTIC.value:
                client = create_elasticsearch_insecure_connection(node)
                if isinstance(client, tuple):
                    raise APIException401(str(client[1]))

                if new_password:
                      res = client.security.change_password(username=user, password=new_password).body
                      if hasattr(res, 'error'):
                          raise APIException403(res.error)

            case DBType.MSSQL.value:
                    client = connect_to_sqlserver_instance(hostname=host.name, uid = node.database_user, pwd=node.database_pwd, prt=node.port)
                    if isinstance(client, tuple):
                        if hasattr(client[1], 'strerror'):
                                raise APIException401(client[1].strerror)
                        raise APIException401(client[1])

                    cursor = client.cursor()
                    if new_password:
                        cursor.execute(f"ALTER LOGIN [{user}] WITH PASSWORD = '{new_password}'")
                        cursor.commit()
    except Exception as e:
        raise APIException(detail = e.strerror if hasattr(e, 'strerror') else str(e), code = e.code if hasattr(e, 'code') else 500)

    return JsonResponse({'notification': f"user {user} updated successfully"}, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DropDatabaseUser(request):
    node_id = request.data.get('node_id')
    user = request.data.get('user')
    node = c123_node.objects.filter(pk = node_id, ind_deleted = False).first()
    if not node:
        raise APIException410(f"db {node_id} is not available")

    host = node.host
    try:
        match node.type:
            case DBType.POSTGRES.value:
                client = get_postgres_connection(node)
                if isinstance(client, tuple):
                    error = client[1].strerror if hasattr(client[1], 'strerror') else str(client[1])
                    raise APIException401(error)

                cursor = client.cursor()
                cursor.execute("DROP USER %s" % user)
                client.commit()
            case DBType.MYSQL.value | DBType.MARIADB.value:
                        client = create_mysql_connection(node)
                        if isinstance(client, tuple):
                            if hasattr(client[1], 'strerror'):
                                raise APIException401(client[1].strerror)
                            raise APIException401(client[1])

                        cursor = client.cursor()
                        cursor.execute("DROP USER `%s`;" % (user,))
                        cursor.execute("FLUSH PRIVILEGES;")
                        client.commit()

            case DBType.MONGO.value:
                client = create_mongo_connection(node)
                if isinstance(client, tuple):
                    if hasattr(client[1], 'strerror'):
                        raise APIException401(client[1].strerror)
                    raise APIException401(client[1])

                client.admin.command("dropUser", user)
            case DBType.ELASTIC.value:
                client = create_elasticsearch_insecure_connection(node)
                if isinstance(client, tuple):
                    raise APIException401(str(client[1]))

                client.security.delete_role(name=user)
                client.security.delete_user(username=user)
            case DBType.MSSQL.value:
                 client = connect_to_sqlserver_instance(hostname=host.name, uid = node.database_user, pwd=node.database_pwd, prt=node.port)
                 if isinstance(client, tuple):
                    if hasattr(client[1], 'strerror'):
                            raise APIException401(client[1].strerror)
                    raise APIException401(client[1])

                 cursor = client.cursor()
                 cursor.execute(f"DROP USER [{user}]")
                 cursor.commit()

    except Exception as e:
        error = e.strerror if hasattr(e, 'strerror') else str(e)
        print("Drop Database User:", error)
        raise APIException(detail = error, code = e.code if hasattr(e, 'code') else 500)

    return JsonResponse({'notification': f"user {user} deleted successfully"}, status = 200)
class ConfigView(generics.GenericAPIView):
    serializer_class = ConfigSerializer
    permission_classes = [IsAuthenticated]
    def get(self, request, *args, **kwargs):
        mariadb_cluster_sizing, mariadb_created = ClusterSizing.objects.get_or_create(
            db_type=DBType.MARIADB.value)
        elasticsearch_cluster_sizing, elasticsearch_created = ClusterSizing.objects.get_or_create(
            db_type=DBType.ELASTIC.value)
        mongo_cluster_sizing, mongo_created =  ClusterSizing.objects.get_or_create(
            db_type=DBType.MONGO.value)
        mysql_cluster_sizing, mysql_created = ClusterSizing.objects.get_or_create(
            db_type=DBType.MYSQL.value)
        oracle_cluster_sizing, oracle_created = OracleClusterSizing.objects.get_or_create(
            db_type=DBType.ORACLE_CDB.value)
        postgres_cluster_sizing, postgres_created = ClusterSizing.objects.get_or_create(
            db_type=DBType.POSTGRES.value)
        redis_cluster_sizing, redis_created = ClusterSizing.objects.get_or_create(
            db_type=DBType.REDIS.value)
        mssql_cluster_sizing, mssql_created = ClusterSizing.objects.get_or_create(
            db_type=DBType.MSSQL.value)


        pg_sys, pg_sys_created = SystemConfig.objects.get_or_create(db_type = DBType.POSTGRES.value)
        mssql_sys, mssql_sys_created = SystemConfig.objects.get_or_create(db_type = DBType.MSSQL.value)
        mariadb_sys, mariadb_sys_created = SystemConfig.objects.get_or_create(db_type = DBType.MARIADB.value)
        mysql_sys, mysql_sys_created = SystemConfig.objects.get_or_create(db_type = DBType.MYSQL.value)
        elasticsearch_sys, elasticsearch_sys_created = SystemConfig.objects.get_or_create(db_type = DBType.ELASTIC.value)
        mongo_sys, mongo_sys_created = SystemConfig.objects.get_or_create(db_type = DBType.MONGO.value)
        redis_sys, redis_sys_created = SystemConfig.objects.get_or_create(db_type = DBType.REDIS.value)
        oracle_sys, oracle_sys_created = OracleSystemConfig.objects.get_or_create()

        current_user = request.user
        users = User.objects.all()
        is_superuser = current_user.groups.filter(name='Administrators').exists() or current_user.is_superuser
        licenses = CustomerKey.objects.exclude(key = 'summary')
        license_summmary = generate_license_summary()
        serializer = self.get_serializer({'oracle_sys': oracle_sys, 'pg_sys': pg_sys, 'mongo_sys': mongo_sys, 'redis_sys': redis_sys, 'mariadb_sys': mariadb_sys, 'mysql_sys': mysql_sys, 'mssql_sys': mssql_sys, 'elasticsearch_sys': elasticsearch_sys, 'mongo_cluster_sizing': mongo_cluster_sizing, 'mariadb_cluster_sizing': mariadb_cluster_sizing, 'elasticsearch_cluster_sizing': elasticsearch_cluster_sizing, 'redis_cluster_sizing': redis_cluster_sizing, 'mysql_cluster_sizing': mysql_cluster_sizing, 'mssql_cluster_sizing': mssql_cluster_sizing, 'postgres_cluster_sizing': postgres_cluster_sizing, 'oracle_cluster_sizing': oracle_cluster_sizing, 'current_user': current_user, 'users': users, 'licenses': licenses, 'license_summary': license_summmary})
        response = serializer.data
        response["current_user"].update({"is_superuser": is_superuser })
        count = 0
        task_name = Task._meta.get_field("displayment")
        for user in users:
            try:
                task = Task.objects.filter(created_by=user).last()
            except:
                pass
            else:
                if(task):
                    response["users"][count].update({'last_activity': task_name.value_from_object(task)})
                else:
                    response["users"][count].update({'last_activity': "Never"})
            count = count + 1

        return Response(response, status=status.HTTP_200_OK)

@api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def GetSystemSettings(request):

    try:
      setting = SystemSettings.objects.first()
      if setting:
          return JsonResponse({'setting value': setting.value}, status=200)
      else:
          return JsonResponse({'error': 'Setting not found'}, status=404)

    except json.JSONDecodeError:
        return JsonResponse({'error': 'Invalid JSON'}, status=400)
    except Exception as e:
        return JsonResponse({'error': str(e)}, status=500)

class NodeDetailsView(generics.RetrieveAPIView):
    serializer_class = c123_NodeSerializer
    queryset = c123_node.objects.all()
    permission_classes =[IsAuthenticated]

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DropIntegratedService(request):
    service_id = request.data.get('service_id')
    attached_nodes = c123_node.objects.filter(ex_params__service_id = service_id, ind_deleted = False)
    if attached_nodes.exists():
        cluster_names =  ',\n'.join(attached_nodes.values_list("cluster__name", flat=True).distinct())
        raise APIException406(f"There are some clusters registered on this service\n.List of affected clusters:\n{cluster_names}")
    IntegratedService.objects.filter(pk = service_id).delete()
    return JsonResponse({}, status = 204)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def EditIntegratedService(request):
    IntegratedService.objects.filter(pk = request.data.get('id')).update(**request.data)
    return JsonResponse({}, status = 205)



@api_view(['DELETE'])
@permission_classes((IsAuthenticated, ))
def DeleteUser(request, pk):
    user = User.objects.filter(pk = pk).first()
    if user:
        superusers = User.objects.filter(is_active = True, is_superuser = True)
        if superusers.exists():
            admin = superusers.first()
            if admin:
                c123_node.objects.filter(created_by = user, ind_deleted = False).update(created_by = admin)
        user.delete()

    return JsonResponse(data = {}, status = 204)
@api_view(['PUT'])
@permission_classes((IsAuthenticated, ))
def UpdateSystemConfig(request):
     db_type = request.data.get("db_type")
     if DBType.ORACLE_CDB.value in db_type or   DBType.ORACLE_PDB.value in db_type:
        config = OracleSystemConfig.objects.first()
        serializer_class = OracleSystemConfigSerializer
     else:
        config = SystemConfig.objects.filter(db_type = db_type).first()
        serializer_class = SystemConfigSerializer
     serializer = serializer_class(config, data = request.data)

     if serializer.is_valid():
        serializer.save()

     return JsonResponse({"notification": "system config successfully updated"}, status = 202)

@api_view(['PUT'])
@permission_classes((IsAuthenticated, ))
def UpdateSystemSettings(request):

    try:
      data = request.data
      setting_name = data.get('name')
      new_value = data.get('value')
      if setting_name is None or new_value is None:
          return JsonResponse({'error': 'Name and value are required'}, status=400)

      setting = SystemSettings.objects.filter(name=setting_name).first()
      if setting:
          setting.value = new_value
          setting.save()
          return JsonResponse({'notification': 'System settings successfully updated'}, status=202)
      else:
          return JsonResponse({'error': 'Setting not found'}, status=404)

    except json.JSONDecodeError:
        return JsonResponse({'error': 'Invalid JSON'}, status=400)
    except Exception as e:
        return JsonResponse({'error': str(e)}, status=500)

class TaskLogLastList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_TaskLogSerializer
    def get_queryset(self):
        last_1_minute = timezone.now() - timezone.timedelta(minutes=1)
        user = self.request.user
        queryset = TaskLog.objects.filter(created_by = user, created_date__gte=last_1_minute, task_id__isnull = False).order_by("-task_id", "-created_date").distinct("task_id")
        id_to_exclude = Task.objects.exclude(monitor_name__isnull = True).exclude(schedule = Schedule.NOW.value).union(Task.objects.filter(task__in=['pg_cluster_primary_search_task', 'pg_split_brain_recovery_task'])).values_list('task_id', flat = True)
        return queryset.exclude(task_id__in=id_to_exclude)[:5]
class DatabaseList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_DatabaseSerializer
    def get_queryset(self):
        user = self.request.user
        if user.is_superuser or user.groups.filter(name='Administrators').exists():
            return c123_database.objects.filter(ind_deleted = False)
        return c123_database.objects.filter(created_by = user, ind_deleted = False)
class HAProxyList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = c123_NodeSerializer
    def get_queryset(self):
        user = self.request.user
        if user.is_superuser or user.groups.filter(name='Administrators').exists():
            return c123_node.objects.filter(type = DBType.HAPROXY.value, ind_deleted = False)

        return c123_node.objects.filter(created_by = user, type = DBType.HAPROXY.value, ind_deleted = False)

class TaskRunningLastList(generics.ListAPIView):
    permission_classes =[IsAuthenticated]
    serializer_class = TaskSerializer
    def get_queryset(self):
        user = self.request.user
        return Task.objects.filter(created_by = user, displayment__isnull = False, status__in=[TaskState.PENDING.value, TaskState.RUNNING.value]).exclude(displayment__exact='').exclude(monitor_name__isnull = False).order_by("-created_date")

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckTaskStatus(request):
    """check status of task"""
    task_id = request.data.get("task_id")
    task = Task.objects.filter(task_id = task_id).first()
    if not task_id:
        raise APIException410(f"Task {task_id} is not available")
    last_step = TaskStepLog.objects.filter(task_id = task_id).last()
    if last_step:
        match last_step.status:
            case Severity.OK.value:
                if last_step.step == 'task finished':
                    task_status = 'completed ok'
                else:
                    # since last step is not 'task finished'
                    task_status = 'in progress'
            case _ :
                # since last step are not green
                task_status = "there are errors"
    else:
        # since we still dont have any step for this task
        task_status = "pending"

    return JsonResponse({"notification": f'task {task_id} ({task.task}) status: {task_status}'}, status=200)


@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DropBackup(request):
    """drop a backup"""
    backup_id = request.data.get("backup_id")
    api_call = request.data.get("rest_api")
    if api_call:
        backup = c123_inventory.objects.filter(name=backup_id)
    else:
        backup = c123_inventory.objects.filter(pk=backup_id)

    if not backup.exists():
        raise APIException404(f"backup id {backup_id} not found")
    task = drop_backup(backup_list = backup)
    return JsonResponse({"notification": f'{task.displayment} started'}, status=200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def ChangeClusterSize(request):
    """change the size scheme for node"""
    new_size = request.data.get("new_size")
    node_id = request.data.get("node_id")
    api_call = request.data.get('rest_api')
    if api_call:
        node = c123_node.objects.filter(name=node_id, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(pk=node_id, ind_deleted = False).first()

    if not node:
        raise APIException410(f"node id {node_id} is not available")
    node.sizing = new_size
    node.save()
    # serialize the objects
    node_data = object_to_json(node)
    host_data = c123_HostSerializer(node.host).data
    created_by = request.user
    if DBType.ORACLE_PDB.value in node.type:
        cdb = c123_node.objects.filter(cluster = node.cluster, type = DBType.ORACLE_CDB.value, cluster_role = ClusterRoles.CDB.value, ind_deleted = False).first()
        if not cdb:
            raise APIException410(f"CDB of PDB {node.name} is not available")
        cdb_data = object_to_json(cdb)
        tmp_data = json.dumps({'host_data': host_data, 'pdb_data': node_data, 'cdb_data': cdb_data, 'new_size': new_size, 'created_by': created_by.id})
        task = Task.objects.create(task='change_pdb_cluster_size_task',
                                displayment='configure {} settings'.format(node.name),
                                target_type=DBType.ORACLE_PDB.value,
                                target_id=node.pk,
                                org_id=node.org_id,
                                schedule=Schedule.NOW.value,
                                status=TaskState.PENDING.value,
                                parameters=tmp_data,
                                created_by=created_by)
    else:
        tmp_data = json.dumps({'host_data': host_data, 'node_data': node_data, 'new_size': new_size, 'created_by': created_by.id})
        task = Task.objects.create(task='change_cluster_size_task',
                                displayment='configure {} settings'.format(node.name),
                                target_type=node.type,
                                target_id=node.pk,
                                org_id=node.org_id,
                                schedule=Schedule.NOW.value,
                                status=TaskState.PENDING.value,
                                parameters=tmp_data,
                                created_by=created_by)
    return JsonResponse({"notification": f"{task.displayment} started"}, status=200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckPortExists(request):
    host_name = request.data.get("host")
    port = request.data.get("port")
    username = request.data.get("os_username")
    password = request.data.get("os_password")
    if password:
        password = format_password(password)
    private_key = request.data.get("private_key")
    ssh_port = request.data.get("ssh_port")
    node_id = request.data.get("node_id")
    type = request.data.get("type")
    node = None
    if node_id:
        node = c123_node.objects.filter(pk = node_id).first()
        if not node:
            raise APIException404(f'host node id {node_id} is not found')

        host = node.host
    else:
         host = SimpleNamespace( **{
        "name": host_name,
        "username": username,
        "password": password,
        "private_key": private_key,
        "ssh_port": ssh_port
        })

    result = check_port_existence(host, port, type, node)

    return JsonResponse({"exists": result}, status = 200)


@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckK8sControlPlaneStatus(request):
    token = request.data.get("api_token")
    hostname = request.data.get("api_host")
    port = request.data.get("api_port")

    url = f"https://{hostname}:{port}/healthz"
    headers = {"Authorization": token}

    try:
        response = requests.get(url, headers=headers, timeout=5, verify=False)
        # Raise HTTPError for bad responses (4xx, 5xx)
        response.raise_for_status()
        return JsonResponse({"status": "available", "message": "Control plane is operational"}, status=200)
    except requests.exceptions.RequestException as e:
        return JsonResponse({"status": "unavailable", "error": str(e)}, status=503)


@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckProxmoxConnection(request):
    try:
        hostname = request.data.get("hostname")
        proxmox_username = request.data.get("username")
        proxmox_token_name = request.data.get("proxmox_token_name")
        proxmox_token_value = request.data.get("proxmox_token_value")
        verify_ssl = request.data.get("verify_ssl")
        exists = test_proxmox_connection(hostname=hostname,
                                         proxmox_username=proxmox_username,
                                         proxmox_token_name=proxmox_token_name,
                                         proxmox_token_value=proxmox_token_value,
                                         verify_ssl=verify_ssl)
        return JsonResponse({"exists": exists}, status = 200)

    except Exception as e:
        raise APIException400(str(e))

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckAWSConnection(request):
    try:
        username = request.data.get("username")
        password = request.data.get("password")
        exists = True
        from clusterCore.AWS import AWSController
        try:
          AWSController(aws_access_key=username, aws_secret_key=password).getAllRegions()
        except Exception as e:
            print(str(e))
            exists = False

        return JsonResponse({"exists": exists}, status = 200)

    except Exception as e:
        raise APIException400(str(e))

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetSubnets(request):
    try:
        username = request.data.get("username")
        password = request.data.get("password")
        region = request.data.get("region")
        from clusterCore.AWS import AWSController
        subnets = AWSController(username, password).getVPCSubnets(region)
        return JsonResponse({"subnets": subnets}, status = 200)

    except Exception as e:
        raise APIException500(str(e))

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckConnection(request):
    host_name = request.data.get("host")
    username = request.data.get("os_username")
    password = request.data.get("os_password")
    if password:
        password = format_password(password)
    private_key = request.data.get("private_key")
    port = request.data.get("ssh_port")
    db_type = request.data.get("db_type")
    db_port = request.data.get("db_port")
    db_pwd =  request.data.get("db_pwd")
    if db_pwd:
        db_pwd = format_password(db_pwd)
    db_user = request.data.get("db_user")
    if db_type and DBType.MSSQL.value in db_type:
        node = SimpleNamespace(**{"database_user":db_user, "database_pwd": db_pwd, "port": db_port, "host": SimpleNamespace(**{"name": host_name})})
        client = connect_to_sqlserver_instance(node)
    else:
        client = create_ssh_connection(host_name, username, password, private_key, port)

    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
                raise APIException401(client[1].strerror)
        raise APIException401(client[1])

    if db_type == DBType.MSSQL.value:
        client = create_winrm_connection(hostname=host_name, user=username, pwd=password, prt=port)
        if isinstance(client, tuple):
            if hasattr(client[1], 'strerror'):
                    raise APIException401(client[1].strerror)
            raise APIException401(client[1])


    return JsonResponse({"connected": True}, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckPathExists(request):
    host_name = request.data.get("host")
    path = request.data.get("path")
    username = request.data.get("os_username")
    password = request.data.get("os_password")
    if password:
        password = format_password(password)
    private_key = request.data.get("private_key")
    node_id = request.data.get("node_id")
    ssh_port = request.data.get("ssh_port")
    host_type = request.data.get("host_type")
    if node_id:
        node = c123_node.objects.filter(pk = node_id).first()
        if not node:
            raise APIException404(f'host node id {node_id} is not found')

        host = node.host
    else:
         host = SimpleNamespace( **{
        "name": host_name,
        "username": username,
        "password": password,
        "private_key": private_key,
        "ssh_port": ssh_port
        })


    result = checkPathExists(host, path, host_type)


    return JsonResponse({"exists": result}, status = 200)
@api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def GetNodeDatabases(request, id):
    node = c123_node.objects.filter(pk = id, ind_deleted = False).first()
    if not node:
        raise APIException410(f" node id {id} is not available")

    current_user = request.user

    if node.type == DBType.ORACLE_CDB.value:
        objects = c123_node.objects.filter(type = DBType.ORACLE_PDB.value, cluster = node.cluster).distinct("name").values("pk", "name")
    else:
        objects = c123_database.objects.filter(type = node.type, node = node).distinct("name").values("pk", "name")

    app_set = set(o["name"] for o in objects)
    match node.type:
        case DBType.ORACLE_CDB.value:
            try:
                conn = connect_to_oracle_database(cdb=node)
            except Exception as e:
                error, = e.args
                raise APIException401(str(error))
            try:
                cursor = conn.cursor()
                cursor.execute(f"""SELECT p.PDB_NAME "Pluggable database name",
DECODE(SUM(d.BYTES), NULL, 'Not Available since the PDB is down', (SUM(d.BYTES))) "Size(B)"
FROM DBA_PDBS p, CDB_DATA_FILES d
WHERE p.PDB_ID=d.CON_ID(+) AND p.PDB_NAME NOT LIKE 'PDB$SEED'
GROUP BY p.PDB_ID, p.PDB_NAME
ORDER BY SUM(d.BYTES) DESC""")
                real_data = {r[0]: float(r[1]) if 'is down' not in r[1] else 0.0 for r in cursor.fetchall()}
                real_set = set(k for k in real_data.keys())
            except Exception as error:
                print(str(error))
                raise APIException403(str(error))

        case DBType.MYSQL.value | DBType.MARIADB.value:
            conn = create_mysql_connection(node)
            if isinstance(conn, tuple):
                if hasattr(conn[1], 'strerror'):
                    raise APIException401(conn[1].strerror)
                raise APIException401(conn[1])

            try:
                cursor = conn.cursor(dictionary=True)
                cursor.execute("SELECT schema_name AS DatabaseName, ROUND((SELECT SUM(data_length + index_length) / 1024 FROM information_schema.tables WHERE table_schema = schemata.schema_name), 2) AS SizeKB FROM information_schema.schemata WHERE schema_name NOT IN ('mysql', 'information_schema', 'performance_schema', 'sys', 'test', 'tmp');")
                real_data = {r['DatabaseName']: r['SizeKB'] for r in cursor.fetchall()}
                real_set = set(k for k in real_data.keys())
            except Exception as error:
                raise APIException403(str(error))

        case DBType.MONGO.value:
            conn = create_mongo_connection(node)
            if isinstance(conn, tuple):
                if hasattr(conn[1], 'strerror'):
                    raise APIException401(conn[1].strerror)
                raise APIException401(conn[1])


            try:
                real_data = {r['name']: round(r["sizeOnDisk"] / 1024, 2) for r in conn.list_databases() if r['name'] not in ["admin", "local", "config"]}
                real_set = set(k for k in real_data.keys())
            except Exception as error:
                raise APIException403(str(error))

        case DBType.MSSQL.value:
            conn = connect_to_sqlserver_instance(hostname=node.host.name, uid = node.database_user, pwd=node.database_pwd, prt=node.port)
            if isinstance(conn, tuple):
                if hasattr(conn[1], 'strerror'):
                      raise APIException401(conn[1].strerror)
                raise APIException401(conn[1])

            try:
                cursor = conn.cursor()
                cursor.execute("SELECT DB_NAME(mf.database_id) AS DatabaseName, SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE mf.database_id > 4 GROUP BY mf.database_id;")
                real_data = {r.DatabaseName: r.SizeKB for r in cursor.fetchall()}
                real_set = set(k for k in real_data.keys())
            except Exception as error:
                print(error)
                raise APIException403(str(error))

        case DBType.ELASTIC.value:
            conn = create_elasticsearch_insecure_connection(node)
            if isinstance(conn, tuple):
                if hasattr(conn[1], 'strerror'):
                    raise APIException401(conn[1].strerror)
                raise APIException401(conn[1])
            try:
                real_data = {name: round(data["total"]["store"]["size_in_bytes"] / 1024, 2) for name, data in conn.indices.stats(metric="store")["indices"].items() if not name.startswith(".")}
                real_set = set(k for k in real_data.keys())
            except Exception as error:
                raise APIException403(str(error))

        case DBType.POSTGRES.value:
            conn = get_postgres_connection(node)
            if isinstance(conn, tuple):
                error = conn[1].strerror if hasattr(conn[1], 'strerror') else str(conn[1])
                raise APIException401(error)

            try:
                cursor = conn.cursor(cursor_factory=RealDictCursor)
                cursor.execute("SELECT datname AS DatabaseName, ROUND(pg_database_size(datname)::numeric / 1024, 2) AS SizeKB FROM pg_database WHERE datistemplate = false AND datname NOT IN ('postgres', 'repmgr')")
                real_data = {r['databasename']: r['sizekb'] for r in cursor.fetchall()}
                real_set = set(k for k in real_data.keys())
            except Exception as error:
                raise APIException403(str(error))
        case DBType.POSTGRES_K8S.value:
            conn = get_postgres_connection(node)
            if isinstance(conn, tuple):
                error = conn[1].strerror if hasattr(conn[1], 'strerror') else str(conn[1])
                raise APIException401(error)

            try:
                cursor = conn.cursor(cursor_factory=RealDictCursor)
                cursor.execute(
                    "SELECT datname AS DatabaseName, ROUND(pg_database_size(datname)::numeric / 1024, 2) AS SizeKB FROM pg_database WHERE datistemplate = false AND datname NOT IN ('postgres', 'repmgr')")
                real_data = {r['databasename']: r['sizekb'] for r in cursor.fetchall()}
                real_set = set(k for k in real_data.keys())
            except Exception as error:
                raise APIException403(str(error))
    existing = real_set.intersection(app_set)
    gone = app_set - real_set
    new = real_set - app_set
    for name in gone:
        objects.filter(name = name).update(severity = Severity.NOT_OK.value, status = Status.DROPPED.value, ind_deleted = True, size = 0.0)
    for name in existing:
        objects.filter(pk__in = objects.filter(name = name).values_list('pk', flat=True)).update(severity = Severity.OK.value, status = Status.OK.value, ind_deleted = False, size = real_data[name])

    if node.type == DBType.ORACLE_CDB.value:
        for name in new:
            data = {}
            data['name'] = name
            data['status'] = Status.OK.value
            data['cluster'] = node.cluster
            data['created_by'] = current_user
            data['last_update_by'] = current_user
            data['host'] = node.host
            data['type'] = DBType.ORACLE_PDB.value
            data['severity'] = Severity.OK.value
            data['sizing'] = Node_Size_Allowed.MEDIUM.name
            data['size'] = real_data[name]
            data['dir_123Cluster'] = node.dir_123Cluster
            data['ex_params'] = {
            'cdb_id': node.id,
            'also_start_standby': False,
            'always_use_this_location': False,
            'replicate_to_standby': "none",
        }
            objects.create(**data)
        response = c123_NodeSerializer(c123_node.objects.filter(type = DBType.ORACLE_PDB.value, cluster = node.cluster, ind_deleted = False), many = True).data
        for pdb in response:
            pdb["exports"] = c123_inventory_serializer(c123_inventory.objects.filter(type = Inventory.EXPORT.value, node_id = pdb['id']), many = True).data
            pdb["backups"] = c123_inventory_serializer(c123_inventory.objects.filter(type = Inventory.BACKUP.value, node_id = pdb['id']), many = True).data
            pdb["flashbacks"] = c123_inventory_serializer(c123_inventory.objects.filter(type = Inventory.FLASHBACK.value, node_id = pdb['id']), many = True).data
    else:
        for name in new:
             objects.create(name=name,
               type=node.type,
               created_by=current_user,
               last_update_by=current_user,
               status=Status.OK.value,
               severity=Severity.OK.value,
               size = real_data[name],
               database_user=node.database_user,
               database_pwd=node.database_pwd,
               node = node)

        response = c123_DatabaseSerializer(c123_database.objects.filter(type = node.type, node = node), many = True).data
        for database in response:
            database["exports"] = c123_inventory_serializer(c123_inventory.objects.filter(type = Inventory.EXPORT.value, database_id = database['id']), many = True).data


    return JsonResponse(response, safe = False, status = 200)
